# -*- mode: python; coding: utf-8 -*-
# Copyright 2019 the AAS WorldWide Telescope project
# Licensed under the MIT License.

import numpy as np
import pytest

from .. import pyramid
from ..pyramid import Pos


def test_next_highest_power_of_2():
    assert pyramid.next_highest_power_of_2(1) == 256
    assert pyramid.next_highest_power_of_2(256) == 256
    assert pyramid.next_highest_power_of_2(257) == 512


def test_depth2tiles():
    assert pyramid.depth2tiles(0) == 1
    assert pyramid.depth2tiles(1) == 5
    assert pyramid.depth2tiles(2) == 21
    assert pyramid.depth2tiles(10) == 1398101


def test_is_subtile():
    from ..pyramid import is_subtile

    assert is_subtile(Pos(2, 0, 0), Pos(1, 0, 0)) == True

    with pytest.raises(ValueError):
        is_subtile(Pos(1, 0, 0), Pos(2, 0, 0))


def test_pos_parent():
    from ..pyramid import pos_parent

    assert pos_parent(Pos(7, 65, 33)) == (Pos(6, 32, 16), 1, 1)

    with pytest.raises(ValueError):
        pos_parent(Pos(0, 0, 0))


def test_generate_pos():
    from ..pyramid import generate_pos

    assert list(generate_pos(0)) == [Pos(0, 0, 0)]

    assert list(generate_pos(1)) == [
        Pos(1, 0, 0),
        Pos(1, 1, 0),
        Pos(1, 0, 1),
        Pos(1, 1, 1),
        Pos(0, 0, 0),
    ]



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
from numpy import testing as nt
import os.path
import pytest
import sys

from . import assert_xml_elements_equal, test_path
from .. import cli
from .. import study


class TestStudy(object):
    WTML = """<?xml version='1.0' encoding='UTF-8'?>
<Folder Browseable="True" Group="Explorer" MSRCommunityId="0" MSRComponentId="0" Name="Toasty" Permission="0" Searchable="True" Type="Sky">
  <Place Angle="0.0" AngularSize="0.0" DataSetType="Sky" Dec="0.0" Distance="0.0" DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0" MSRCommunityId="0" MSRComponentId="0" Name="Toasty" Opacity="100.0" Permission="0" RA="0.0" Rotation="0.0" Thumbnail="thumb.jpg" ZoomLevel="1.0">
    <ForegroundImageSet>
      <ImageSet BandPass="Visible" BaseDegreesPerTile="1.0" BaseTileLevel="0" BottomsUp="False" CenterX="0.0" CenterY="0.0" DataSetType="Sky" ElevationModel="False" FileType=".png" Generic="False" MeanRadius="0.0" MSRCommunityId="0" MSRComponentId="0" Name="Toasty" OffsetX="0.0" OffsetY="0.0" Permission="0" Projection="Tan" Rotation="0.0" Sparse="True" StockSet="False" TileLevels="4" Url="{1}/{3}/{3}_{2}.png" WidthFactor="2">
        <ThumbnailUrl>thumb.jpg</ThumbnailUrl>
      </ImageSet>
    </ForegroundImageSet>
  </Place>
</Folder>
"""

    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_basic(self):
        tiling = study.StudyTiling(2048, 2048)
        assert tiling._width == 2048
        assert tiling._height == 2048
        assert tiling._p2n == 2048
        assert tiling._tile_size == 8
        assert tiling._tile_levels == 3
        assert tiling._img_gx0 == 0
        assert tiling._img_gy0 == 0


    def test_preconditions(self):
        with pytest.raises(ValueError):
            study.StudyTiling(0, 1)

        with pytest.raises(ValueError):
            study.StudyTiling(1, -1)

        with pytest.raises(ValueError):
            study.StudyTiling(1, np.nan)


    def test_image_to_tile(self):
        tiling = study.StudyTiling(514, 514)
        assert tiling._p2n == 1024
        assert tiling.image_to_tile(0, 0) == (0, 0, 255, 255)
        assert tiling.image_to_tile(0, 513) == (0, 3, 255, 0)
        assert tiling.image_to_tile(513, 0) == (3, 0, 0, 255)
        assert tiling.image_to_tile(513, 513) == (3, 3, 0, 0)


    def test_sample_cli(self):
        from xml.etree import ElementTree as etree
        expected = etree.fromstring(self.WTML)

        for variants in ([], ['--placeholder-thumbnail']):
            args = ['tile-study']
            args += variants
            args += [
                '--outdir', self.work_path(),
                test_path('NGC253ALMA.jpg')
            ]
            cli.entrypoint(args)

        with open(self.work_path('index_rel.wtml')) as f:
            observed = etree.fromstring(f.read())

        assert_xml_elements_equal(observed, expected)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Loading OpenEXR files.

This is very primitive support. Implemented for:

https://svs.gsfc.nasa.gov/4851

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
load_openexr
'''.split()

import numpy as np
import sys


def _util_load_openexr_float(path):
    """
    A diagnostic/utility function to load OpenEXR as float data.

    """
    import OpenEXR
    import Imath

    EXR_TO_NUMPY = {
        Imath.PixelType.FLOAT: np.float32,
        Imath.PixelType.HALF: np.float16,
    }

    exr = OpenEXR.InputFile(path)
    header = exr.header()
    dw = header['dataWindow']
    width = dw.max.x - dw.min.x + 1
    height = dw.max.y - dw.min.y + 1

    img = None

    for idx, chan in enumerate('RGB'):
        ctype = header['channels'][chan].type
        cbytes = exr.channel(chan)
        dtype = EXR_TO_NUMPY[ctype.v]

        if img is None:
            img = np.empty((height, width, 3), dtype=dtype)

        img[...,idx] = np.frombuffer(cbytes, dtype=dtype).reshape((height, width))

    return img


def load_openexr(path):
    """
    Load an OpenEXR file

    Parameters
    ----------
    path : path-like
        The path to the file

    Returns
    -------
    An image-like Numpy array with shape ``(height, width, planes)``
    and a dtype of uint8.

    """
    try:
        import OpenEXR
        import Imath
    except ImportError as e:
        raise Exception('cannot load OpenEXR file: needed support libraries are not available') from e

    EXR_TO_NUMPY = {
        Imath.PixelType.FLOAT: np.float32,
        Imath.PixelType.HALF: np.float16,
    }

    exr = OpenEXR.InputFile(path)
    header = exr.header()
    dw = header['dataWindow']
    width = dw.max.x - dw.min.x + 1
    height = dw.max.y - dw.min.y + 1

    if header['lineOrder'] != Imath.LineOrder(Imath.LineOrder.INCREASING_Y):
        raise Exception('cannot load OpenEXR file: unsupported lineOrder')
    if len(header['channels']) != 3:
        raise Exception('cannot load OpenEXR file: expected exactly 3 channels')
    if 'chromaticities' in header:
        print('warning: ignoring chromaticities in OpenEXR file; colors will be distorted',
              file=sys.stderr)
    if 'whiteLuminance' in header:
        print('warning: ignoring whiteLuminance in OpenEXR file; colors will be distorted',
              file=sys.stderr)

    img = np.empty((height, width, 3), dtype=np.uint8)

    try:
        for idx, chan in enumerate('RGB'):
            ctype = header['channels'][chan].type
            cbytes = exr.channel(chan)
            carr = np.frombuffer(cbytes, dtype=EXR_TO_NUMPY[ctype.v]).reshape((height, width))

            # XXX: manually implemented sRGB conversion is not awesome, and also
            # ideally this wouldn't be done here. Equations from
            # https://discourse.techart.online/t/converting-linear-exr-to-srgb-jpeg-with-python/2267/3

            # Need a read-write buffer:
            work = carr.copy()

            # Small-value branch of sGRB. Assume that there aren't many of
            # these, and avoid modifying non-small values.
            mask = (carr <= 0.0031308)
            del carr
            work[mask] *= 12.92 * 255
            np.copyto(img[...,idx], work, where=mask, casting='unsafe')

            # Main branch. Replace small values with something that won't
            # give math errors.
            work[mask] = 0.0032
            np.power(work, 1 / 2.4, out=work)
            np.multiply(work, 1.055, out=work)
            np.subtract(work, 0.055, out=work)
            np.multiply(work, 255, out=work)
            np.logical_not(mask, out=mask)
            np.copyto(img[...,idx], work, where=mask, casting='unsafe')
    except Exception as e:
        raise Exception('cannot load OpenEXR file: unexpected file structure') from e

    return img


# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2020 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os
from xml.dom.minidom import parseString
from tempfile import mkstemp, mkdtemp
from shutil import rmtree

import pytest
import numpy as np

try:
    import healpy as hp
    from astropy.io import fits
    HAS_ASTRO = True
except ImportError:
    HAS_ASTRO = False

try:
    import OpenEXR
    HAS_OPENEXR = True
except ImportError:
    HAS_OPENEXR = False

from . import test_path
from .. import toast
from ..image import ImageLoader, ImageMode
from ..pyramid import Pos, PyramidIO
from ..samplers import plate_carree_sampler, healpix_fits_file_sampler
from ..toast import sample_layer


def test_mid():
    from .._libtoasty import mid

    result = mid((0, 0), (np.pi / 2, 0))
    expected = np.pi / 4, 0
    np.testing.assert_array_almost_equal(result, expected)

    result = mid((0, 0), (0, 1))
    expected = 0, .5
    np.testing.assert_array_almost_equal(result, expected)


def test_area():
    MAX_DEPTH = 6
    areas = {}

    for t in toast.generate_tiles(MAX_DEPTH, bottom_only=False):
        a = areas.get(t.pos.n, 0)
        areas[t.pos.n] = a + toast.toast_tile_area(t)

    for d in range(1, MAX_DEPTH + 1):
        np.testing.assert_almost_equal(areas[d], 4 * np.pi)


def test_tile_for_point_boundaries():
    # These test points occur at large-scale tile boundaries and so we're not
    # picky about where they land in the tiling -- either tile on a border is
    # OK.

    from ..toast import toast_tile_for_point

    latlons = [
        (0., 0.),

        (0., -0.5 * np.pi),
        (0., 0.5 * np.pi),
        (0., np.pi),
        (0., 1.5 * np.pi),
        (0., 2 * np.pi),
        (0., 2.5 * np.pi),

        (-0.5 * np.pi, 0.),
        (-0.5 * np.pi, -np.pi),
        (-0.5 * np.pi, np.pi),

        (0.5 * np.pi, 0.),
        (0.5 * np.pi, -np.pi),
        (0.5 * np.pi, np.pi),
    ]

    for depth in range(4):
        for lat, lon in latlons:
            tile = toast_tile_for_point(depth, lat, lon)


def test_tile_for_point_specifics():
    from ..toast import toast_tile_for_point

    test_data = {
        (0.1, 0.1): [
            (0, 0, 0),
            (1, 1, 0),
            (2, 3, 1),
            (3, 7, 3),
            (4, 14, 7),
            (5, 29, 14),
            (6, 59, 29),
            (7, 119, 59),
            (8, 239, 119),
            (9, 479, 239),
            (10, 959, 479),
            (11, 1918, 959),
            (12, 3837, 1918),
        ]
    }

    for (lat, lon), nxys in test_data.items():
        for nxy in nxys:
            tile = toast_tile_for_point(nxy[0], lat, lon)
            assert tile.pos.n == nxy[0]
            assert tile.pos.x == nxy[1]
            assert tile.pos.y == nxy[2]


def test_pixel_for_point():
    from ..toast import toast_pixel_for_point

    test_data = {
        (0.1, 0.1): (3, 7, 3, 126, 191),
        (-0.4, 2.2): (3, 1, 1, 200, 75),
    }

    for (lat, lon), (n, x, y, px, py) in test_data.items():
        shallow_tile, frac_x, frac_y = toast_pixel_for_point(n, lat, lon)
        assert shallow_tile.pos.n == n
        assert shallow_tile.pos.x == x
        assert shallow_tile.pos.y == y
        assert px == int(round(frac_x))
        assert py == int(round(frac_y))
        deep_tile, _, _ = toast_pixel_for_point(n + 8, lat, lon)
        assert deep_tile.pos.x % 256 == px
        assert deep_tile.pos.y % 256 == py


def image_test(expected, actual, err_msg):
    resid = np.abs(1. * actual - expected)
    if np.median(resid) < 15:
        return

    _, pth = mkstemp(suffix='.png')
    import PIL.Image
    PIL.Image.fromarray(np.hstack((expected, actual))).save(pth)
    pytest.fail("%s. Saved to %s" % (err_msg, pth))


class TestSampleLayer(object):
    def setup_method(self, method):
        self.base = mkdtemp()
        self.pio = PyramidIO(self.base)

    def teardown_method(self, method):
        rmtree(self.base)

    def verify_level1(self, mode):
        for n, x, y in [(1, 0, 0), (1, 0, 1), (1, 1, 0), (1, 1, 1)]:
            ref_path = test_path('earth_toasted_sky', str(n), str(y), "%i_%i.png" % (y, x))
            expected = ImageLoader().load_path(ref_path).asarray()
            if mode == ImageMode.F32:
                expected = expected.mean(axis=2)

            pos = Pos(n=n, x=x, y=y)
            observed = self.pio.read_toasty_image(pos, mode).asarray()

            image_test(expected, observed, 'Failed for %s' % ref_path)

    def test_plate_carree(self):
        img = ImageLoader().load_path(test_path('Equirectangular_projection_SW-tweaked.jpg'))
        sampler = plate_carree_sampler(img.asarray())
        sample_layer(self.pio, ImageMode.RGB, sampler, 1)
        self.verify_level1(ImageMode.RGB)

    @pytest.mark.skipif('not HAS_OPENEXR')
    def test_earth_plate_caree_exr(self):
        img = ImageLoader().load_path(test_path('Equirectangular_projection_SW-tweaked.exr'))
        sampler = plate_carree_sampler(img.asarray())
        sample_layer(self.pio, ImageMode.RGB, sampler, 1)
        self.verify_level1(ImageMode.RGB)

    @pytest.mark.skipif('not HAS_ASTRO')
    def test_healpix_equ(self):
        sampler = healpix_fits_file_sampler(test_path('earth_healpix_equ.fits'))
        sample_layer(self.pio, ImageMode.F32, sampler, 1)
        self.verify_level1(ImageMode.F32)

    @pytest.mark.skipif('not HAS_ASTRO')
    def test_healpix_gal(self):
        sampler = healpix_fits_file_sampler(test_path('earth_healpix_gal.fits'))
        sample_layer(self.pio, ImageMode.F32, sampler, 1)
        self.verify_level1(ImageMode.F32)
